﻿using static kMeans.Helper;

namespace kMeans;

public static class Program
{
    private const string Coordinates = "coordinates.csv";
    private const int K = 3;
    private const string PlotOut = "plot.png";

    public static void Main()
    {
        Console.WriteLine("k-Means-Algorithm");
        var points = ParseCsv(Coordinates).ToList();

        var centroids = InitializeRandomlyCentroids(points, K).ToList();
        var update = true;
        while (update)
        {
            AssignPointsToCentroids(points, centroids);
            update = UpdateCentroids(points, centroids);
        }

        var plot = CreatePlot(points, centroids);
        ExportAndPreviewPlot(plot, PlotOut);
    }

    private static void AssignPointsToCentroids(List<Point> points, List<Point> centroids)
    {
        foreach (var point in points)
        {
            var id = 0;
            var distance = double.MaxValue;
            foreach (var centroid in centroids)
            {
                // calculate euclid distance and assign id.
                var currentDistance = Distance(centroid, point);
                if (currentDistance > distance) continue;
                distance = currentDistance;
                id = centroid.ClusterId;
            }

            point.ClusterId = id;
        }
    }

    private static double Distance(Point centroid, Point point)
    {
        return Math.Sqrt(Math.Pow(centroid.X - point.X, 2) + Math.Pow(centroid.Y - point.Y, 2));
    }

    private static IEnumerable<Point> InitializeRandomlyCentroids(IReadOnlyCollection<Point> points, int k)
    {
        var minX = points.Min(p => p.X);
        var maxX = points.Max(p => p.X);
        var minY = points.Min(p => p.Y);
        var maxY = points.Max(p => p.Y);
        var rnd = new Random();

        for (var i = 0; i < k; i++)
        {
            var x = (minX + maxX) * rnd.NextDouble();
            var y = (minY + maxY) * rnd.NextDouble();
            var point = new Point(x, y, i);
            Console.WriteLine(point);
            yield return point;
        }
    }

    private static bool UpdateCentroids(IReadOnlyCollection<Point> points, List<Point> centroids)
    {
        // calculate mean of all points of one cluster.
        var updated = false;
        foreach (var centroid in centroids)
        {
            updated |= UpdateClusterCentroid(centroid, points);
        }

        return updated;
    }

    private static bool UpdateClusterCentroid(Point centroid, IEnumerable<Point> points)
    {
        var pointsOfCluster = points.Where(p => p.ClusterId == centroid.ClusterId).ToArray();
        var sumX = pointsOfCluster.Sum(p => p.X);
        var meanX = sumX / pointsOfCluster.Length;
        var sumY = pointsOfCluster.Sum(p => p.Y);
        var meanY = sumY / pointsOfCluster.Length;

        if (!(Math.Abs(centroid.X - meanX) > 1e-5) || !(Math.Abs(centroid.Y - meanY) > 1e-5)) return false;
        centroid.X = meanX;
        centroid.Y = meanY;
        return true;
    }
}